from sqldiff.meta.column_description import ColumnDescription
from sqldiff.meta.postgres_meta import get_meta
from sqldiff.meta.teradata_meta import get_meta

from sqldiff.comp.compare import compare, ColumnsComparisonResult, KeyColumnComparisonResult

from sqldiff.relation.relation_manager import get_relations

# launch relation registering
from sqldiff.relation import relation_register



from typing import List

from sqldiff.meta.dbsystem import POSTGRES
from sqldiff.meta.column_description import ColumnDescription
from sqldiff.meta.meta_connection_dispatcher import dbmeta

LIMIT_0 = ' limit 0'
NUMERIC_MAX_PRECISION = 65535
NUMERIC_MAX_SCALE = 65535

ALIASES = (
    ('SMALLINT', 'INT2'),
    ('INTEGER', 'INT4', 'INT'),
    ('BIGINT', 'INT8'),
    ('NUMERIC', 'DECIMAL'),
    ('REAL', 'FLOAT4'),
    ('DOUBLE', 'DOUBLE PRECISION', 'FLOAT8'),
    ('SMALLSERIAL', 'SERIAL2'),
    ('SERIAL', 'SERIAL4'),
    ('BIGSERIAL', 'SERIAL8'),
    ('VARCHAR', 'CHARACTER VARYING'),
    ('CHAR', 'CHARACTER', 'BPCHAR'),
    ('TIMESTAMP WITHOUT TIME ZONE', 'TIMESTAMP'),
    ('TIMESTAMP WITH TIME ZONE', 'TIMESTAMPTZ'),
    ('TIME WITHOUT TIME ZONE', 'TIME'),
    ('TIME WITH TIME ZONE', 'TIMETZ'),
    ('BOOLEAN', 'BOOL')
)


def get_type_code_name_mapping(connection):
    with connection.cursor() as cursor:
        query = """
        SELECT
        t.oid,
        t.typname
        FROM pg_catalog.pg_type t
        WHERE
        (
            t.typrelid = 0
            OR (
                SELECT c.relkind = 'c'
                FROM pg_catalog.pg_class c
                WHERE c.oid = t.typrelid
            )
        )
        AND NOT EXISTS (SELECT 1 FROM pg_catalog.pg_type el WHERE el.oid = t.typelem AND el.typarray = t.oid)
        AND pg_catalog.pg_type_is_visible(t.oid)
        ORDER BY t.oid
        """
        cursor.execute(query)
        records = cursor.fetchall()
        oid_types_dict = {r[0]: r[1] for r in records}
        return oid_types_dict


@dbmeta('psycopg2')
def get_meta(connection, query, predefined_type_mapping=None):
    if predefined_type_mapping is None:
        type_mapping = get_type_code_name_mapping(connection)
    else:
        type_mapping = predefined_type_mapping

    limit_query = query + LIMIT_0 if not query.strip().endswith(LIMIT_0) else query
    with connection.cursor() as cursor:
        cursor.execute(limit_query)
        metadata: List[ColumnDescription] = []
        position_index = 1
        for column in cursor.description:
            type_name = type_mapping[column.type_code].upper()

            if 'CHAR' in type_name:
                precision = column.internal_size
                scale = None
            else:
                precision = column.precision
                scale = column.scale
            column_description = {
                'name': column.name.lower(),
                'type_code': column.type_code,
                'display_size': column.display_size,
                'internal_size': column.internal_size,
                'precision': precision,
                'scale': scale,
                'null_ok': column.null_ok,
                'type_name': type_name,
                'position': position_index,
                'dbsystem': POSTGRES
            }
            metadata.append(ColumnDescription(**column_description))
            position_index += 1
        return metadata



import importlib
import sys
from collections import namedtuple

import pkg_resources
from packaging import version

meta_registry = []

ConnectionMeta = namedtuple('ConnectionMeta', 'driver_name min_version max_version meta_func')


def dbmeta(driver_name, min_version=None, max_version=None):
    def decorator(func):
        meta_registry.append(ConnectionMeta(driver_name, min_version, max_version, func))
        return func
    return decorator


def check_obj(obj):
    # inspec.get()
    mod = importlib.util.module_from_spec(obj)
    base, _sep, _stem = mod.__name__.partition('.')
    return sys.modules[base]


# https://stackoverflow.com/questions/2020014/get-fully-qualified-class-name-of-an-object-in-python
def fullname(o):
    # o.__module__ + "." + o.__class__.__qualname__ is an example in
    # this context of H.L. Mencken's "neat, plausible, and wrong."
    # Python makes no guarantees as to whether the __module__ special
    # attribute is defined, so we take a more circumspect approach.
    # Alas, the module name is explicitly excluded from __qualname__
    # in Python 3.
    module = o.__class__.__module__
    if module is None or module==str.__class__.__module__:
        return o.__class__.__name__  # Avoid reporting __builtin__
    else:
        return module + '.' + o.__class__.__name__


def get_meta_provider(connection_obj):
    connection_fullname = fullname(connection_obj)
    pkg_name = connection_fullname.split('.')[0]
    pkg_version = pkg_resources.get_distribution(pkg_name).version
    for i in meta_registry:
        if pkg_name==i.driver_name and (
                not i.min_version or version.parse(pkg_version) >= version.parse(i.min_version)) and (
                not i.max_version or version.parse(pkg_version) <= version.parse(i.max_version)):
            return i.meta_func
    raise TypeError('No metadata provider defined for provided connection.')



import functools


class ColumnDescription:
    """
    PEP 249 - https://www.python.org/dev/peps/pep-0249/
    DB API Cursor description provides sequence of 7 items describing one result column:
    name, type_code, display_size, internal_size, precision, scale, null_ok
    The first two items (name and type_code) are mandatory,
    the other five are optional and are set to None if no meaningful values can be provided.

    Thi class contains additional fields such as:
    source database type (dbsystem),
    type_name (not provided in cursor desciption),
    position
    """

    def __init__(self, name, type_code, display_size, internal_size,
                 precision, scale, null_ok, type_name, position, dbsystem):
        self.name = name
        self.type_code = type_code
        self.display_size = display_size
        self.internal_size = internal_size
        self.precision = precision
        self.scale = scale
        self.null_ok = null_ok
        self.type_name = type_name
        self.position = position
        self.dbsystem = dbsystem

    def __repr__(self):
        column_desc = f"ColumnDescription({self.position=}, {self.dbsystem=}, {self.name=}, {self.type_name=}, " \
                      f"{self.precision=}, {self.scale=}, {self.internal_size=}, " \
                      f"{self.null_ok=}, {self.display_size=}, {self.type_code=})"
        column_desc = column_desc.replace('self.', '')
        return column_desc

    def __str__(self):
        s_prec = str(self.precision) if self.precision is not None else ''
        s_scale = str(self.scale) if self.scale is not None else ''
        s = f'{self.name} {self.type_name}({", ".join(filter(None, (s_prec, s_scale)))})'.replace('()', '')
        return s

    @functools.cached_property
    def dict(self):
        return {
            self.name: self
        }


"""
This modules uses https://github.com/andialbrecht/sqlparse for parsing and analyzing sql statements.

"""

import re
from functools import wraps

import sqlparse as sqlparse


def is_schema_dot_table_string(db_object_string: str):
    pattern = re.compile('^(\w+\.\w+|\w+)$')
    return bool(pattern.match(db_object_string))


def schema_table_name_to_query(tab_name):
    if is_schema_dot_table_string(tab_name):
        return f'select * from {tab_name};'
    raise ValueError("Provided value must be in format 'schema.table'")


# Any function using this decorator assumes that it'll get a sql string or parsed query as parameter
# If many_Statements_allowed == False, then function expects parsed statement instead of a tuple of parsed statements
def parse_sql(many_statements_allowed=False):
    """
    Parse sql statement using sqlparser if str is provided as function parameter.
    :param many_statements_allowed: Default False.
    If False - raises exception when parsed query has more than one statement.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(sql):
            if isinstance(sql, str):
                parsed = sqlparse.parse(sql)
            else:
                parsed = sql
            if len(parsed) > 1 and not many_statements_allowed:
                raise ValueError('Sql string or parsed query is not allowed to have more than one sql statement.')
            elif len(parsed) > 1 and many_statements_allowed:
                return func(parsed)
            elif len(parsed) == 1 and not many_statements_allowed:
                return func(parsed[0])
            else:
                return func(parsed)
        return wrapper
    return decorator

# Todo fix parser (dont work when parsed is empty string)
@parse_sql()
def is_select_statement(parsed):
    return parsed.get_type() == 'SELECT'





# TODO DEFINE RELATIONS
from copy import deepcopy

from sqldiff.meta import dbsystem
from sqldiff.comp.compare import ColumnDescription
from sqldiff.relation.relation_manager import dbmapping


@dbmapping('DECIMAL', dbsystem.TERADATA)
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'NUMERIC'
    return tgt_col


@dbmapping('TIMESTAMP WITH TIME ZONE', dbsystem.TERADATA)
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'TIMESTAMPTZ'
    tgt_col.precision = None
    tgt_col.scale = None
    return tgt_col


@dbmapping('TIME WITH TIME ZONE', dbsystem.TERADATA)
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'TIMESTZ'
    tgt_col.precision = None
    tgt_col.scale = None
    return tgt_col


@dbmapping('CHAR', dbsystem.TERADATA)
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'VARCHAR'
    return tgt_col


@dbmapping('CHAR', dbsystem.TERADATA, (dbsystem.POSTGRES,))
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'BPCHAR'
    return tgt_col


@dbmapping('LONG VARCHAR', dbsystem.TERADATA)
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'VARCHAR'
    tgt_col.precision = 64000
    return tgt_col


@dbmapping('LONG VARCHAR', dbsystem.TERADATA, (dbsystem.POSTGRES,))
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'VARCHAR'
    tgt_col.precision = -1
    return tgt_col


@dbmapping('VARCHAR', dbsystem.TERADATA, (dbsystem.POSTGRES,))
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'VARCHAR'
    tgt_col.precision = -1
    return tgt_col


@dbmapping('CHAR', dbsystem.TERADATA, (dbsystem.POSTGRES,))
def teradata_long_varchar_mapping(src_col: ColumnDescription):
    tgt_col = deepcopy(src_col)
    tgt_col.type_name = 'VARCHAR'
    tgt_col.precision = -1
    return tgt_col





from operator import itemgetter, attrgetter
from typing import List

from sqldiff.comp.syntax_analysis import is_schema_dot_table_string, schema_table_name_to_query, is_select_statement
from sqldiff.meta.column_description import ColumnDescription
from sqldiff.meta.meta_connection_dispatcher import get_meta_provider
from sqldiff.relation.relation_manager import get_relations


KEYS_ORDER_TARGET = 'target'
KEYS_ORDER_SOURCE = 'source'
KEYS_ORDER_ZIP = 'zip'
KEY_ORDER_OPTIONS = (KEYS_ORDER_TARGET, KEYS_ORDER_SOURCE, KEYS_ORDER_ZIP)


def get_item_or_none(iterable, i):
    try:
        return iterable[i]
    except IndexError:
        return None


def compare_keys_zip(source_keys, target_keys):
    pos_factor = 0
    src_idx = 0
    tgt_idx = 0
    key_comp_list = []
    # set of common keys
    common_keys = source_keys.keys() & target_keys.keys()

    common_keys_upper = set([k.upper() for k in source_keys.keys()]) & set([k.upper() for k in target_keys.keys()])

    while src_idx < len(source_keys) or tgt_idx < len(target_keys):

        src_key = get_item_or_none(list(source_keys.keys()), src_idx)
        tgt_key = get_item_or_none(list(target_keys.keys()), tgt_idx)

        src_key_upper = src_key.upper() if src_key is not None else None
        tgt_key_upper = tgt_key.upper() if tgt_key is not None else None

        if src_key_upper in common_keys_upper and tgt_key_upper in common_keys_upper:
            comparison_dict = {
                'source': source_keys[src_key],
                'target': target_keys[tgt_key],
                'exists_on_source': True,
                'exists_on_target': True,
                'match': True,  # exists_on_source and exists_on_target
                'inserted_on_source': False,
                'inserted_on_target': False,
                'order_match': bool(src_idx + pos_factor==tgt_idx and src_key_upper==tgt_key_upper)
            }
            key_comp_list.append(comparison_dict)
            src_idx += 1
            tgt_idx += 1

        else:

            if tgt_key_upper not in common_keys_upper and tgt_key_upper is not None:
                comparison_dict = {
                    'source': None,
                    'target': target_keys[tgt_key],
                    'exists_on_source': False,
                    'exists_on_target': True,
                    'match': False,
                    'inserted_on_source': False,
                    'inserted_on_target': bool(src_idx + pos_factor==tgt_idx),
                    'order_match': bool(src_idx + pos_factor==tgt_idx)
                }
                key_comp_list.append(comparison_dict)
                tgt_idx += 1
                pos_factor += 1

            if src_key_upper not in common_keys_upper and src_key_upper is not None:
                comparison_dict = {
                    'source': source_keys[src_key],
                    'target': None,
                    'exists_on_source': True,
                    'exists_on_target': False,
                    'match': False,
                    'inserted_on_source': bool(src_idx + pos_factor==tgt_idx),
                    'inserted_on_target': False,
                    'order_match': bool(src_idx + pos_factor==tgt_idx)
                }
                key_comp_list.append(comparison_dict)
                src_idx += 1
                pos_factor -= 1

    return key_comp_list


class KeysComparisonResult:
    """Stores data about comparison result of 2 lists of keys"""

    def __init__(self, source, target, exists_on_source, exists_on_target, match, inserted_on_source,
                 inserted_on_target, order_match):
        self.source = source
        self.target = target
        self.exists_on_source = exists_on_source
        self.exists_on_target = exists_on_target
        self.match = match
        self.inserted_on_source = inserted_on_source
        self.inserted_on_target = inserted_on_target
        self.order_match = order_match

    def __repr__(self):
        return f'({self.source}, {self.target})'


def compare_keys(source, target, key=None) -> List[KeysComparisonResult]:
    getter = key if key is not None else lambda a: a
    # dict key insertion order
    source_keys = {getter(i): i for i in source}
    target_keys = {getter(i): i for i in target}

    res = compare_keys_zip(source_keys, target_keys)
    keys_comparison = [KeysComparisonResult(**r) for r in res]

    return keys_comparison


class AttributesComparisonResult:
    COMPARISON_RESULT_SCORECARD = {
        'precision_match': lambda x: (not x) * 5,
        'precision_higher_on_target': lambda x: x * 1,
        'precision_higher_on_source': lambda x: x * 2,
        'scale_match': lambda x: (not x) * 5,
        'scale_higher_on_target': lambda x: x * 1,
        'scale_higher_on_source': lambda x: x * 2,
        'type_name_match': lambda x: (not x) * 50
    }

    def __init__(self, source: ColumnDescription, target: ColumnDescription, relation=None):
        self.relation = relation
        self.source = source
        self.target = target

        # calculate comparison result flags
        precision_is_comparable = source.precision is not None and target.precision is not None
        scale_is_comparable = source.scale is not None and target.scale is not None
        self.precision_match = source.precision==target.precision if precision_is_comparable else True
        self.precision_higher_on_target = source.precision < target.precision if precision_is_comparable else False
        self.precision_higher_on_source = source.precision > target.precision if precision_is_comparable else False
        self.scale_match = source.scale==target.scale if scale_is_comparable else True
        self.scale_higher_on_target = source.scale < target.scale if scale_is_comparable else False
        self.scale_higher_on_source = source.scale > target.scale if scale_is_comparable else False
        self.type_name_match = source.type_name==target.type_name

        self.scale_factor = target.scale - source.scale if scale_is_comparable else 0
        self.precision_factor = target.precision - source.precision if precision_is_comparable else 0

        # count comparison score
        self.scorecard = self.COMPARISON_RESULT_SCORECARD
        self.score = self.calculate_score()

        self.match = self.score==0

    def calculate_score(self):
        score = sum([self.scorecard[k](v) for k, v in vars(self).items() if k in self.scorecard])
        return score

    def set_score(self):
        self.score = self.calculate_score()

    def __repr__(self):
        return f'({self.match=}, {self.type_name_match=}, {self.scale_match=}, {self.precision_match=})'


class ColumnComparisonResult:

    def __init__(self, source: ColumnDescription, target: ColumnDescription):
        # store source and tet columns data
        self.source = source
        self.target = target

        # Get all mapping functions for source column to target database system
        relations = get_relations(src_col=source, tgt_col=target)
        # Map source column using every returned relation
        mapped_source_columns = [rel(source) for rel in relations]
        # Compare mapped fields with target field
        self.attr_comp_results = [
            AttributesComparisonResult(src_mapped_column, target)
            for src_mapped_column, relation in zip(mapped_source_columns, relations)
        ]
        self.attr_comp_results.sort(key=attrgetter('score'))

        self.match = self.attr_comp_results[0].match

    def best_result(self) -> AttributesComparisonResult:
        return self.attr_comp_results[0]

    def __repr__(self):
        if self.attr_comp_results:
            return repr(self.best_result())
        return 'N/A'


class KeyColumnComparisonResult:
    def __init__(self, key: KeysComparisonResult, col: ColumnComparisonResult):
        self.key = key
        self.col = col
        self.match = key.match and col.match

        attr = self.col.best_result() if self.col is not None else None

        self.precision_match = attr.precision_match if attr is not None else None
        self.sacle_match = attr.scale_match if attr is not None else None
        self.type_name_match = attr.type_name_match if attr is not None else None
        self.precision_factor = attr.precision_factor if attr is not None else None
        self.scale_factor = attr.scale_factor if attr is not None else None

    def __repr__(self):
        return f'{repr(self.key)} - {repr(self.col)}'


class ColumnsComparisonResult:
    """
    This class stores compare_column_lists result [Dict] and provides some methods for analyzing comparison result.
    """

    def __init__(self, source_columns: List[ColumnDescription], target_columns: List[ColumnDescription]):
        self.source_columns = source_columns
        self.target_columns = target_columns

        self.src_cols_dict = {col_desc.name: col_desc for col_desc in source_columns}
        self.tgt_cols_dict = {col_desc.name: col_desc for col_desc in target_columns}

        self.keys_comparison = compare_keys(source_columns, target_columns, key=attrgetter('name'))

        # columns comparison - using target column for comparing attributes for matching columns
        self.columns_comparison = [
            ColumnComparisonResult(self.src_cols_dict[kc.target.name], self.tgt_cols_dict[kc.target.name])
            if kc.match else None
            for kc in self.keys_comparison
        ]

        self.kc_result = [
            KeyColumnComparisonResult(k, c)
            for k, c in zip(self.keys_comparison, self.columns_comparison)]

    def __iter__(self):
        return iter(self.kc_result)

    def __len__(self):
        return len(self.keys_comparison)

    def __str__(self):
        max_src = self.get_column_str_max_len('source')
        max_tgt = self.get_column_str_max_len('target')

        header = 'SOURCE TARGET MATCH ORDER_MATCH TYPE_MATCH ATTR_MATCH'.split()
        row_format = '{:>' + str(max_src + 3) + '} | {:' + str(max_tgt + 3) + '} - {:^20}|{:^20}|{:^20}|{:^20}|'
        rows = []
        for kc in self:
            key = kc.key
            source = key.source  # if key is not None else None
            target = key.target  # if key is not None else None
            match = kc.match
            order_match = key.order_match
            type_match = kc.type_name_match
            attr_match = kc.precision_match and kc.sacle_match

            r = map(str, [source, target, match, order_match, type_match, attr_match])

            rows.append(r)
        output_rows = [row_format.format(*header)] + [row_format.format(*r) for r in rows]
        output_string = '\n'.join(output_rows)
        # [row_format.format(header)]# +
        return output_string

    def __repr__(self):
        return str([repr(kc) for kc in self])


    def get_column(self, col_name, col_source='target'):
        """
        Search and return specified column from comparison_result.

        :param col_name: Column name to look for.
        :param col_source: Specify if column will be looked for in 'source' or 'target' columns.
        :return: comparison_result row with specified column.
        """

        try:
            for c in self.comparison_result:
                if c[col_source].name==col_name:
                    return c
        except KeyError:
            raise KeyError("col_source must have value of 'source' or 'target'.")

    def match(self, order=True):
        """
        Checks if target columns match source columns in terms of names, orders and columns attributes.
        :param order: Default True. Set to False if want to ommit order checking.
        :return: True/False
        """
        # Catching KeyError when 'attributes_comparison_results' doesn't exist (no matching field on source/target)
        try:
            for kc in self:
                if not (kc.key.match and (kc.key.order_match or not order) and kc.col.match):
                    return False
        except KeyError:
            return False
        return True

    def get_column_str_max_len(self, col_source='target'):
        """
        Gets longest column string representation on 'source' or 'target'
        :param col_source: 'source' or 'target'
        :return: length of longest column string representation
        """
        if col_source=='target':
            col_list = self.target_columns
        elif col_source=='source':
            col_list = self.source_columns
        else:
            raise KeyError("col_source must have value of 'source' or 'target'.")
        return max([len(str(c)) for c in col_list])

    def order_by_source(self):
        """
        Returns new object with soruce and target fields ordered by source fields
        :return:
        """
        inf = len(self.source_columns) + len(self.target_columns)
        source_order = {c.name: i for i, c in enumerate(self.source_columns)}
        target_order = {c.name: source_order.get(c.name, inf+i) for i, c in enumerate(self.target_columns)}
        target_columns_ordered_by_source = sorted(self.target_columns, key=lambda x:target_order[x.name])
        return ColumnsComparisonResult(self.source_columns, target_columns_ordered_by_source)

    def order_by_target(self):
        """
        Returns new object with soruce and target fields ordered by target fields
        :return:
        """
        inf = len(self.source_columns) + len(self.target_columns)
        target_order = {c.name: i for i, c in enumerate(self.target_columns)}
        source_order = {c.name: target_order.get(c.name, inf+i) for i, c in enumerate(self.source_columns)}
        source_columns_ordered_by_target = sorted(self.source_columns, key=lambda x:source_order[x.name])
        return ColumnsComparisonResult(source_columns_ordered_by_target, self.target_columns)


# TODO change compare source / target functions to "order" funcns
# zip will be the basic for sorting
# for sorting keygetter will be used to get zip string, and string with field name will be mapped to position on source/target


# def _compare_keys_source(source_keys, target_keys):
#     l1 = [(src_key, src_key if src_key in target_keys else None) for src_key in source_keys.keys()]
#     l2 = [(None, tgt_key) for tgt_key in target_keys if tgt_key not in source_keys]
#     l3 = l1 + l2
#     a = 1


def validate_compare_query(query):
    if is_schema_dot_table_string(query):
        return schema_table_name_to_query(query)
    if is_select_statement(query):
        return query
    else:
        raise ValueError('Invalid statement. Must be schema.table or SELECT statement.')


def compare(source_connection, source_query, target_connection, target_query,
            source_get_meta=None, target_get_meta=None):
    # TODO - metadata provider dispatching based on connection object
    # validsate query - check if it is quyery or schema.table, check if theres only one query, chec if it is select query.
    src_sql = validate_compare_query(source_query)
    tgt_sql = validate_compare_query(target_query)

    source_meta_func = source_get_meta if source_get_meta is not None else get_meta_provider(source_connection)
    target_meta_func = target_get_meta if target_get_meta is not None else get_meta_provider(target_connection)

    src_cols = source_meta_func(source_connection, src_sql)
    tgt_cols = target_meta_func(target_connection, tgt_sql)

    result = ColumnsComparisonResult(src_cols, tgt_cols)
    return result



import json
from typing import List

from sqldiff.meta.dbsystem import TERADATA
from sqldiff.meta.column_description import ColumnDescription
from sqldiff.meta.meta_connection_dispatcher import dbmeta


def get_raw_meta(connection, query):
    # based on: https://github.com/Teradata/python-driver/blob/master/samples/MetadataFromPrepare.py
    with connection.cursor() as cursor:
        metadata_query = '{fn teradata_rpo(S)}{fn teradata_fake_result_sets}' + query
        cursor.execute(metadata_query)
        row = cursor.fetchone()
        return json.loads(row[7])


@dbmeta('teradatasql')
def get_meta(connection, query):
    raw_meta = get_raw_meta(connection, query)
    metadata: List[ColumnDescription] = []
    for column in raw_meta:
        type_name = column['TypeName'].upper()

        if 'CHAR' in type_name:
            precision = column['MaxCharacterCount']
            scale = None
        elif 'DECIMAL' == type_name:
            precision = column['Precision']
            scale = column['Scale']
        elif any(i in type_name for i in ('TIME', 'INTERVAL')):
            precision = None
            scale = column['Scale']
        else:
            precision = scale = None

        column_description = {
            'name': column['Title'].lower(),
            'type_code': column['RawDataType'],
            'display_size': None,
            'internal_size': column['ByteCount'],
            'precision': precision,
            'scale': scale,
            'null_ok': column['MayBeNull'],
            'type_name': type_name,
            'position': column['Position'],
            'dbsystem': TERADATA
        }
        metadata.append(ColumnDescription(**column_description))
    return metadata




import unittest
from copy import deepcopy
from operator import itemgetter

from sqldiff.comp.compare import compare_keys, ColumnsComparisonResult, AttributesComparisonResult, \
    ColumnComparisonResult
from sqldiff.meta.column_description import ColumnDescription
from sqldiff.relation.relation_manager import dbmapping, mappings


def create_column(dbsystem, field_name, type_name, precision=None, scale=None):
    """Shortcut function creating ColumnDescription instances for testing purposes"""
    col = ColumnDescription(
        name=field_name,
        type_code=None,
        display_size=None,
        internal_size=None,
        precision=precision,
        scale=scale,
        null_ok=None,
        type_name=type_name,
        position=None,
        dbsystem=dbsystem
    )
    return col


class TestColumnDescriptionStringRepresentation(unittest.TestCase):
    """Test if ColumnDescription __str__ method generates per SQL representation of field"""
    def test_str_column_no_attributes(self):
        col = create_column('dbsystem', 'field_name', 'INT')
        self.assertEqual(str(col), 'field_name INT')

    def test_str_column_precision_only(self):
        col = create_column('dbsystem', 'field_name', 'VARCHAR', 50)
        self.assertEqual(str(col), 'field_name VARCHAR(50)')

    def test_str_column_precision_and_scale(self):
        col = create_column('dbsystem', 'field_name', 'DECIMAL', 19, 5)
        self.assertEqual(str(col), 'field_name DECIMAL(19, 5)')


class TestCompareKeys(unittest.TestCase):
    """Test keys (field names) comparison"""

    def test_compare_keys_unequal_field_lists(self):
        """Test if _compare_keys_zip functions handles unequal fields number"""
        src_fields = ['f1', 'f2', 'f3', 'f5', 'f6']
        tgt_fields = ['f1', 'f3', 'f2', 'f4', 'f5', 'f6']

        result = compare_keys(src_fields, tgt_fields)
        result_fields = [(i.source, i.target) for i in result]
        expected_fields = [
            ('f1', 'f1'),
            ('f2', 'f3'),
            ('f3', 'f2'),
            (None, 'f4'),
            ('f5', 'f5'),
            ('f6', 'f6')
        ]

        self.assertEqual(result_fields, expected_fields)

    def test_compare_keys_unequal_field_lists_different_character_case(self):
        """Test if _compare_keys_zip functions handles unequal fields number"""
        src_fields = ['f1', 'f2', 'f3', 'F5', 'F6']
        tgt_fields = ['f1', 'F3', 'f2', 'f4', 'F5', 'f6']

        result = compare_keys(src_fields, tgt_fields)
        result_fields = [(i.source, i.target) for i in result]
        expected_fields = [
            ('f1', 'f1'),
            ('f2', 'F3'),
            ('f3', 'f2'),
            (None, 'f4'),
            ('F5', 'F5'),
            ('F6', 'f6')
        ]

        self.assertEqual(result_fields, expected_fields)

    def test_compare_keys_order_order_flag(self):
        """Test if compare keys result order flag is set correctly"""
        src_fields = ['f1', 'f2', 'f3', 'f5', 'f6']
        tgt_fields = ['f1', 'f3', 'f2', 'f4', 'f5', 'f6']

        result = compare_keys(src_fields, tgt_fields)
        result_fields = [(i.source, i.target, i.order_match) for i in result]
        expected_fields = [
            ('f1', 'f1', True),
            ('f2', 'f3', False),
            ('f3', 'f2', False),
            (None, 'f4', True),
            ('f5', 'f5', True),
            ('f6', 'f6', True)
        ]

        self.assertEqual(result_fields, expected_fields)

    def test_compare_ks_using_custom_attrgetter(self):
        """Test compare keys on complex structure using optional 'key' argument to pass attrgetter"""
        src_fields = [
            {
                'field_name': 'f1',
                'type_name': 'some_type'
            },
            {
                'field_name': 'f2',
                'type_name': 'some_type'
            },
            {
                'field_name': 'f3',
                'type_name': 'some_type'
            }
        ]
        tgt_fields = [
            {
                'field_name': 'f1',
                'type_name': 'some_type'
            },
            {
                'field_name': 'f3',
                'type_name': 'some_type'
            }
        ]

        getter = itemgetter('field_name')
        result = compare_keys(src_fields, tgt_fields, key=getter)

        result_fields = [
            (
                getter(i.source) if i.source is not None else None,
                getter(i.target) if i.target is not None else None
            ) for i in result
        ]
        expected_fields = [
            ('f1', 'f1'),
            ('f2', None),
            ('f3', 'f3'),
        ]

        self.assertEqual(result_fields, expected_fields)


class TestCompareColumnsAttributes(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def test_compare_column_attributes(self):
        """Test compare columntributes"""
        src_col = create_column('dbsystem1', 'field_name', 'DECIMAL', 19, 9)
        tgt_col = create_column('dbsystem2', 'field_name', 'DECIMAL', 19, 5)

        result = AttributesComparisonResult(src_col, tgt_col)

        self.assertTrue(result.precision_match)
        self.assertFalse(result.precision_higher_on_target)
        self.assertFalse(result.precision_higher_on_source)
        self.assertFalse(result.scale_match)
        self.assertFalse(result.scale_higher_on_target)
        self.assertTrue(result.scale_higher_on_source)
        self.assertTrue(result.type_name_match)

    def test_comparison_result_score_calculation(self):
        """Test comparisonm result score calculation"""

        src_col = create_column('dbsystem1', 'field_name', 'DECIMAL', 19, 9)
        tgt_col = create_column('dbsystem2', 'field_name', 'DECIMAL', 19, 5)

        attr = AttributesComparisonResult(src_col, tgt_col)

        attr.precision_match = True
        attr.precision_higher_on_target = False
        attr.precision_higher_on_source = False
        attr.scale_match = False
        attr.scale_higher_on_target = False
        attr.scale_higher_on_source = True
        attr.type_name_match = True

        attr.scorecard = {
            'precision_match': lambda x: (not x) * 5,
            'precision_higher_on_target': lambda x: x * 1,
            'precision_higher_on_source': lambda x: x * 2,
            'scale_match': lambda x: (not x) * 5,
            'scale_higher_on_target': lambda x: x * 1,
            'scale_higher_on_source': lambda x: x * 2,
            'type_name_match': lambda x: (not x) * 50
        }

        score_result = attr.calculate_score()
        self.assertEqual(score_result, 7)


class TestCompareColumnsWithoutMappings(unittest.TestCase):

    def setUp(self) -> None:
        # set up dbmappings
        # setup columns
        pass

    def test_compare_the_same_columns(self):
        """Test if comparing the same columns returns perfect match"""
        src_col = create_column('dbsystem1', 'field_name', 'DECIMAL', 19, 5)
        tgt_col = create_column('dbsystem1', 'field_name', 'DECIMAL', 19, 5)

        result = ColumnComparisonResult(src_col, tgt_col)
        # result.kc_result.

        self.assertTrue(result.best_result().precision_match)
        self.assertTrue(result.best_result().scale_match)
        self.assertTrue(result.best_result().type_name_match)

    def test_compare_different_unmapped_types(self):
        """Test if comparing the same columns returns perfect match"""
        src_col = create_column('dbsystem1', 'field_name', 'DECIMAL', 19, 5)
        tgt_col = create_column('dbsystem1', 'field_name', 'VARCHAR', 50)

        result = ColumnComparisonResult(src_col, tgt_col)

        self.assertFalse(result.best_result().precision_match)
        # scale is not comparable
        self.assertTrue(result.best_result().scale_match)
        self.assertFalse(result.best_result().type_name_match)


class CompareColumnsWithCustomMappings(unittest.TestCase):
    """Test compare_columns function with custom dbmappings registered"""

    def setUp(self) -> None:
        # TODO fix dbmapping duplication on multiple tup call
        @dbmapping('LONG VARCHAR', 'teradata')
        def teradata_long_varchar_varchar_mapping(src_col: ColumnDescription):
            ret_col = deepcopy(src_col)
            ret_col.type_name = 'VARCHAR'
            ret_col.precision = 64000
            return ret_col

        @dbmapping('LONG VARCHAR', 'teradata', 'postgres')
        def teradata_long_varchar_text_postgres_mapping(src_col: ColumnDescription):
            ret_col = deepcopy(src_col)
            ret_col.type_name = 'TEXT'
            ret_col.precision = None
            ret_col.scale = None
            return ret_col

    def test_compare_columns_matching_mapping(self):
        src_col = create_column('teradata', 'field_name', 'LONG VARCHAR')

        tgt_col = create_column('postgres', 'field_name', 'VARCHAR', 64000)
        result = ColumnComparisonResult(src_col, tgt_col)
        self.assertEqual(result.best_result().score, 0)

        tgt_col = create_column('postgres', 'field_name', 'TEXT')
        result = ColumnComparisonResult(src_col, tgt_col)
        self.assertEqual(result.best_result().score, 0)

    def test_compare_columns_not_matching_mapping(self):
        m = mappings
        src_col = create_column('teradata', 'field_name', 'LONG VARCHAR')
        tgt_col = create_column('postgres', 'field_name', 'DECIMAL', 19, 5)
        result = ColumnComparisonResult(src_col, tgt_col)
        self.assertNotEqual(result.best_result().score, 0)


class TestCompareColumnLists(unittest.TestCase):

    def test_compare_column_lists_matching_with_order(self):
        src_cols = [
            create_column('teradata', 'f1', 'DECIMAL', 19, 5),
            create_column('teradata', 'f2', 'VARCHAR', 500),
            create_column('teradata', 'f3', 'DATE')
        ]

        tgt_cols = [
            create_column('teradata', 'f1', 'DECIMAL', 19, 5),
            create_column('teradata', 'f2', 'VARCHAR', 500),
            create_column('teradata', 'f3', 'DATE')
        ]
        # Matching type but not the same
        result = ColumnsComparisonResult(src_cols, tgt_cols)

        self.assertTrue(result.match())
        self.assertTrue(result.match(order=False))

    def test_compare_column_lists_matching_wrong_column_order(self):
        src_cols = [
            create_column('teradata', 'f1', 'DECIMAL', 19, 5),
            create_column('teradata', 'f2', 'VARCHAR', 500),
            create_column('teradata', 'f3', 'DATE')
        ]

        tgt_cols = [
            create_column('teradata', 'f1', 'DECIMAL', 19, 5),
            create_column('teradata', 'f3', 'DATE'),
            create_column('teradata', 'f2', 'VARCHAR', 500)
        ]
        # Matching type but not the same
        result = ColumnsComparisonResult(src_cols, tgt_cols)

        self.assertFalse(result.match())
        self.assertTrue(result.match(order=False))

    def test_compare_not_matching_column_lists(self):
        src_cols = [
            create_column('teradata', 'f1', 'DECIMAL', 19, 5),
            create_column('teradata', 'f2', 'VARCHAR', 500),
            create_column('teradata', 'f3', 'DATE'),
            create_column('teradata', 'f5', 'VARCHAR', 50),
            create_column('teradata', 'f6', 'DATE'),
            create_column('teradata', 'f7', 'VARCHAR', 100),

        ]

        tgt_cols = [
            create_column('teradata', 'f1', 'DECIMAL', 19, 9),
            create_column('teradata', 'f3', 'DATE'),
            create_column('teradata', 'f2', 'VARCHAR', 100),
            create_column('teradata', 'f5', 'VARCHAR', 50),
            create_column('teradata', 'f9', 'INT'),
            create_column('teradata', 'f6', 'TIMESTAMP'),
            create_column('teradata', 'f4', 'INT'),

        ]
        result = ColumnsComparisonResult(src_cols, tgt_cols)

        # Todo some more specific tests on comparison attributes
        print(result)
        self.assertFalse(result.match())

    def test_ordered_comparison_result(self):
        """Test source ordered comparison"""
        src_cols = [
            create_column('teradata', 'f1', 'DECIMAL', 19, 5),
            create_column('teradata', 'f2', 'VARCHAR', 500),
            create_column('teradata', 'f3', 'DATE'),
            create_column('teradata', 'f5', 'VARCHAR', 50),
            create_column('teradata', 'f6', 'DATE'),
            create_column('teradata', 'f7', 'VARCHAR', 100),

        ]

        tgt_cols = [
            create_column('teradata', 'f1', 'DECIMAL', 19, 9),
            create_column('teradata', 'f3', 'DATE'),
            create_column('teradata', 'f2', 'VARCHAR', 100),
            create_column('teradata', 'f5', 'VARCHAR', 50),
            create_column('teradata', 'f9', 'INT'),
            create_column('teradata', 'f6', 'TIMESTAMP'),
            create_column('teradata', 'f4', 'INT'),

        ]
        result = ColumnsComparisonResult(src_cols, tgt_cols)

        source_ordered = result.order_by_source()
        target_ordered = result.order_by_target()
        a = 1



from copy import deepcopy

import sqldiff.meta.dbsystem as dbsystem
from sqldiff.meta import ColumnDescription


from sqldiff.utils import make_tuple

ANY_DB = (dbsystem.POSTGRES, dbsystem.TERADATA)

# Dictionary which holds relation functions registered by dbmapping decorator
mappings = {}


def dbmapping(type_name, source_dbsystem, target_dbsystem=ANY_DB):
    """
    Decorator adds defined function to registry of dbsystems types mappings.
    :param type_name:
    :param source_dbsystem:
    :param target_dbsystem:
    :return:
    """

    def decorator(func):
        # make systems iterable - in case that single string, or other sequence type has been passed as an argument
        source_dbsystems = make_tuple(source_dbsystem)
        target_dbsystems = make_tuple(target_dbsystem)

        for src_db in source_dbsystems:
            for tgt_db in target_dbsystems:
                key = (type_name, src_db, tgt_db)
                mapping = mappings.setdefault(key, [])
                if func not in mapping:
                    mapping.append(func)
        return func
    return decorator


def relation_identity(src_col: ColumnDescription):
    res_col = deepcopy(src_col)
    return res_col


def get_relations_using_keys(type_name, source_dbsystem, target_dbsystem):
    if not all([isinstance(i, str) for i in (type_name, source_dbsystem, target_dbsystem)]):
        raise ValueError('Relation keys needs to be strings')
    return mappings.get((type_name, source_dbsystem, target_dbsystem), []) + [relation_identity]


def get_relations_using_columns(src_col: ColumnDescription, tgt_col: ColumnDescription):
    if not all([isinstance(i, ColumnDescription) for i in (src_col, tgt_col)]):
        raise ValueError('Parameters needs to be ColumnDescription type')
    return get_relations_using_keys(src_col.type_name, src_col.dbsystem, tgt_col.dbsystem)


def get_relations(**kwargs):
    """
    Returns all registered relations either for:
    str kwargs: ('type_name', 'source_dbsystem', 'target_dbsystem')
    or ColumnDescription kwargs: ('src_col', 'tgt_col')
    :param kwargs:
    :return:
    """
    if all(i in kwargs for i in ('src_col', 'tgt_col')):
        return get_relations_using_columns(**kwargs)
    if all(i in kwargs for i in ('type_name', 'source_dbsystem', 'target_dbsystem')):
        return get_relations_using_keys(**kwargs)
    raise KeyError('Invalid parameters')


# czy dany (typ, dbsystem) mapuje sie na (typ, dbsystem) ????



import setuptools

setuptools.setup(
    name='sqldiff',
    packages=setuptools.find_packages(),
    version='0.0.4',
    license='MIT',
    description='Compare sql objects structure (tables, views or queries) on different databases',
    author='Mateusz Matelski',
    author_email='m.z.matelski@gmail.com',
    url='https://github.com/m-matelski/sqldiff',
    download_url = 'https://github.com/m-matelski/sqldiff/archive/v0.0.4.tar.gz',
    keywords = ['sql', 'diff', 'postgres', 'teradata'],
    install_requires=[
        'sqlparse==0.4.1', 'packaging>=20.4'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
)


import unittest

import sqlparse

from sqldiff.comp.syntax_analysis import is_select_statement


class TestSyntax(unittest.TestCase):

    def test1(self):
        q = "select * from a;" \
            "select * from b;"
        parsed = sqlparse.parse(q)
        a = 1

    def test_is_select_statement(self):
        q = "select * from a"
        a = is_select_statement(q)
        b = 1


def make_tuple(obj):
    """
    Converts param object to tuple considering if the provided object is iterable.
    :param obj: Any iterable or non-iterable argument
    :return: tuple(obj) if param is iterable, (param,) otherwise [one element tuple with obj param as an element]
    """
    try:
        iter(obj)
        if isinstance(obj, str):
            obj_tuple = (obj,)
        else:
            obj_tuple = tuple(obj)
    except TypeError:
        obj_tuple = (obj,)
    return obj_tuple




import unittest

from sqldiff.meta.column_description import ColumnDescription
from sqldiff.relation.relation_manager import dbmapping, mappings, get_relations_using_keys, get_relations_using_columns


class TestRelationManagerDbmappingDecorator(unittest.TestCase):
    """
    Test if @dbmapping decorator stores mapping fuions properly.
    Mapping registration shouldn't be imported into current module, so mapping dictionary is empty.
    """

    def test_single_dbsystem_string_registering(self):
        """Test if dbmapping decorator handles single dbsystems values"""

        type_name = 'type_name'
        src_dbsystem = 'src_dbsystem'
        tgt_dbsystem = 'tgt_dbsystem'

        @dbmapping(type_name, src_dbsystem, tgt_dbsystem)
        def test_function():
            pass

        self.assertIn(test_function, mappings[(type_name, src_dbsystem, tgt_dbsystem)])

    def test_multiple_dbsystem_string_registering(self):
        """Test if decorator handles dbsystems values as sequences"""

        type_name = 'type_name'
        src_dbsystems = ('src_dbsystem1', 'src_dbsystem2')
        tgt_dbsystems = ('tgt_dbsystem1', 'tgt_dbsystem2')

        @dbmapping(type_name, src_dbsystems, tgt_dbsystems)
        def test_function():
            pass

        for src_dbsystem in src_dbsystems:
            for tgt_dbsystem in tgt_dbsystems:
                self.assertIn(test_function, mappings[(type_name, src_dbsystem, tgt_dbsystem)])

    def test_get_relations_using_keys(self):
        """Test if getting relations by keys works, and returns additional idety relation function"""

        type_name = 'type_name'
        src_dbsystem = 'src_dbsystem'
        tgt_dbsystem = 'tgt_dbsystem'

        @dbmapping(type_name, src_dbsystem, tgt_dbsystem)
        def test_function():
            pass

        rel = get_relations_using_keys(type_name, src_dbsystem, tgt_dbsystem)

        self.assertEqual(len(rel), 2)


    def test_get_relations_using_keys(self):
        """Test if getting relations by ColumnDescription, and returns additional idety relation function"""

        src_col = ColumnDescription(
            name='src',
            type_code=None,
            display_size=None,
            internal_size=None,
            precision=None,
            scale=None,
            null_ok=None,
            type_name='some type',
            position=None,
            dbsystem='some db system'
        )
        tgt_col = ColumnDescription(
            name='tgt',
            type_code=None,
            display_size=None,
            internal_size=None,
            precision=None,
            scale=None,
            null_ok=None,
            type_name='some type',
            position=None,
            dbsystem='some db system2'
        )

        type_name = 'type_name'
        src_dbsystem = 'src_dbsystem'
        tgt_dbsystem = 'tgt_dbsystem'

        @dbmapping(src_col.type_name, src_col.dbsystem, tgt_col.dbsystem)
        def test_function():
            pass

        rel = get_relations_using_columns(src_col, tgt_col)

        self.assertEqual(len(rel), 2)







